#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Relevant Utilities.

File: utils.py
Author: huxuan
Email: i(at)huxuan.org
"""
import json
import logging
import socket
import struct
from subprocess import PIPE  # noqa: S404
from subprocess import Popen  # noqa: S404
from subprocess import TimeoutExpired  # noqa: S404
from urllib.parse import urlparse

import requests

from iptvtools.config import Config


PROBE_COMMAND = (
    'ffprobe -hide_banner -show_streams -select_streams v '
    '-of json=c=1 -v quiet'
)

UDP_SCHEME = (
    'udp',
    'rtp',
)


def convert_url_with_udpxy(orig_url, udpxy):
    """Convert url with udpxy."""
    parsed_url = urlparse(orig_url)
    if parsed_url.scheme in UDP_SCHEME:
        return f'{udpxy}/{parsed_url.scheme}/{parsed_url.netloc}'
    return orig_url


def unify_title_and_id(item):
    """Unify title and id."""
    for title_unifier in sorted(Config.title_unifiers):
        if title_unifier in item['title']:
            item['title'] = item['title'].replace(
                title_unifier,
                Config.title_unifiers[title_unifier])

    if 'tvg-name' in item.get('params'):
        item['id'] = item['params']['tvg-name']
    else:
        item['id'] = item['title']

    for id_unifier in sorted(Config.id_unifiers):
        if id_unifier in item['id']:
            item['id'] = item['id'].replace(
                id_unifier,
                Config.id_unifiers[id_unifier])

    return item


def probe(url, timeout=None):
    """Invoke probe to get stream information."""
    outs = None
    with Popen(  # noqa: S603
            f'{PROBE_COMMAND} {url}'.split(),
            stdout=PIPE, stderr=PIPE) as proc:
        try:
            outs, _ = proc.communicate(timeout=timeout)
        except TimeoutExpired:
            proc.kill()
    if outs:
        try:
            return json.loads(outs.decode('utf-8'))
        except json.JSONDecodeError as exc:
            logging.error(exc)
    return None


def check_stream(url, timeout=None):
    """Check stream information and return height."""
    stream_info = probe(url, timeout)
    if stream_info and stream_info.get('streams'):
        return max([
            stream.get('height', 0)
            for stream in stream_info['streams']
        ])
    return 0


def check_connectivity(url, timeout=None):
    """Check connectivity."""
    parsed_url = urlparse(url)
    if parsed_url.scheme in UDP_SCHEME:
        return check_udp_connectivity(parsed_url.netloc, timeout)
    return check_http_connectivity(url, timeout)


def check_udp_connectivity(url, timeout=None):
    """Check UDP connectivity."""
    ipaddr, port = url.rsplit(':', 1)
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    sock.settimeout(timeout)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    sock.bind(('', int(port)))
    mreq = struct.pack('4sl', socket.inet_aton(ipaddr), socket.INADDR_ANY)
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
    try:
        if sock.recv(10240):
            return True
    except socket.timeout:
        pass
    return False


def check_http_connectivity(url, timeout=None):
    """Check HTTP connectivity."""
    try:
        return requests.get(url, timeout=timeout, stream=True).ok
    except requests.RequestException:
        return False


def height_to_resolution(height):
    """Convert height to resolution."""
    if not height:
        return ''
    if height >= 4320:
        return '8K'
    if height >= 2160:
        return '4K'
    if height >= 1080:
        return '1080p'
    if height >= 720:
        return '720p'
    return f'{height}p'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Helps for iptvtools.

File: constants.py
Author: huxuan
Email: i(at)huxuan.org
"""
CONFIG = (
    'Configuration file to unify title and id.'
)
CHANNEL_EXCLUDE = (
    'Channels to exclude with regex. '
    'Note: Blacklist has higher priority than whitelist.'
)
CHANNEL_INCLUDE = (
    'Channels to include with regex. '
    'Note: Only channels in the whitelist will be included if set.'
)
GROUP_EXCLUDE = (
    'Groups to exclude with regex.'
    'Note: Blacklist has higher priority than whitelist.'
)
GROUP_INCLUDE = (
    'Groups to include with regex.'
    'Note: Only groups in the whitelist will be included if set.'
)
INPUTS = (
    'One or more input m3u playlist files/urls.'
)
INTERVAL = (
    'Interval in seconds between successive fetching requests.'
)
LOG_LEVEL = 'Log level.'
MIN_HEIGHT = (
    'Minimum height/resolution to accept, 0 means no resolution filtering.'
)
OUTPUT = 'Output file name.'
REPLACE_GROUP_BY_SOURCE = (
    'Flag to replace the group title with the source name, where the source '
    'name is the basename of input files/urls without extension.'
)
RESOLUTION_ON_TITLE = (
    'Flag to append resolution such as 8K, 4K, 1080p, 720p to the title.'
)
SORT_KEYS = (
    'List of keys to sort the channels. Valid options currently supported '
    'are `tvg-id`, `height` and `title`.'
)
TEMPLATES = (
    'Template m3u files/urls with well-maintained channel information to '
    'replace the matched entries.'
)
TIMEOUT = (
    'Timeout threshold for fetching request.'
)
UDPXY = (
    'UDP Proxy for certain IPTV channels.'
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Filter IPTV m3u playlists according to customized criteria.

File: main.py
Author: huxuan
Email: i(at)huxuan.org
"""
import argparse
import logging
import shutil

from iptvtools import __version__
from iptvtools import exceptions
from iptvtools.config import Config
from iptvtools.constants import defaults
from iptvtools.constants import helps
from iptvtools.models import Playlist


def parse_args():
    """Arguments Parsers."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--channel-exclude', help=helps.CHANNEL_EXCLUDE)
    parser.add_argument('--channel-include', help=helps.CHANNEL_INCLUDE)
    parser.add_argument('--group-exclude', help=helps.GROUP_EXCLUDE)
    parser.add_argument('--group-include', help=helps.GROUP_INCLUDE)
    parser.add_argument('--min-height', default=defaults.MIN_HEIGHT, type=int,
                        help=helps.MIN_HEIGHT)
    parser.add_argument('-c', '--config', default=defaults.CONFIG,
                        help=helps.CONFIG)
    parser.add_argument('-i', '--inputs', nargs='*', default=defaults.INPUTS,
                        help=helps.INPUTS)
    parser.add_argument('-I', '--interval', default=defaults.INTERVAL,
                        type=int, help=helps.INTERVAL)
    parser.add_argument('-L', '--log-level', default=defaults.LOG_LEVEL,
                        help=helps.LOG_LEVEL)
    parser.add_argument('-o', '--output', default=defaults.OUTPUT,
                        help=helps.OUTPUT)
    parser.add_argument('-r', '--replace-group-by-source', action='store_true',
                        help=helps.REPLACE_GROUP_BY_SOURCE)
    parser.add_argument('-R', '--resolution-on-title', action='store_true',
                        help=helps.RESOLUTION_ON_TITLE)
    parser.add_argument('-s', '--sort-keys', nargs='*',
                        default=defaults.SORT_KEYS, help=helps.SORT_KEYS)
    parser.add_argument('-t', '--templates', nargs='*',
                        default=defaults.TEMPLATES, help=helps.TEMPLATES)
    parser.add_argument('-T', '--timeout', default=defaults.TIMEOUT, type=int,
                        help=helps.TIMEOUT)
    parser.add_argument('-u', '--udpxy', default=defaults.UDPXY,
                        help=helps.UDPXY)
    parser.add_argument('-v', '--version', action='version',
                        version=__version__)
    return parser.parse_args()


def main():
    """Filter m3u playlists."""
    args = parse_args()

    logging.basicConfig(level=args.log_level.upper())

    if args.min_height or args.resolution_on_title:
        if shutil.which('ffprobe') is None:
            raise exceptions.FFmpegNotInstalledError()

    Config.init(args.config)
    playlist = Playlist(args)
    playlist.parse()
    playlist.filter()
    playlist.export()
    if playlist.inaccessible_urls:
        logging.info('Inaccessible Urls:')
        logging.info('\n'.join(sorted(playlist.inaccessible_urls)))
    if playlist.poor_urls:
        logging.info('Poor resolution Urls:')
        logging.info('\n'.join(sorted(playlist.poor_urls)))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom exceptions for iptvtools.

File: exceptions.py
Author: huxuan
Email: i(at)huxuan.org
"""


class BaseCustomException(RuntimeError):
    """Base Custom Exception."""


class FFmpegNotInstalledError(BaseCustomException):
    """Raise when FFmpeg is not installed."""

    def __init__(self):
        """Init for FfmpegNotInstalledError."""
        super().__init__(
            'Need `FFmpeg` for resolution related processing.\n'
            'Please install it according to '
            '`https://www.ffmpeg.org/download.html`.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Playlist which contains all the channels' information.

File: models.py
Author: huxuan
Email: i(at)huxuan.org
"""
import logging
import os.path
import random
import re
import sys
import time

from tqdm import tqdm

from iptvtools import parsers
from iptvtools import utils
from iptvtools.constants import tags


class Playlist():
    """Playlist model."""

    def __init__(self, args):
        """Init for Playlist."""
        self.args = args
        self.data = {}
        self.id_url = {}
        self.inaccessible_urls = set()
        self.poor_urls = set()
        self.tvg_url = None

    def export(self):
        """Export playlist information."""
        res = []
        res.append(tags.M3U)
        if self.tvg_url is not None:
            res[0] += f' x-tvg-url="{self.tvg_url}"'
        for url in sorted(self.data, key=self.__custom_sort):

            if url in self.inaccessible_urls or url in self.poor_urls:
                continue

            entry = self.data[url]
            params_dict = entry.get('params', {})
            if self.args.replace_group_by_source:
                params_dict['group-title'] = self.data[url]['source']
            params = ' '.join([
                f'{key}="{value}"'
                for key, value in params_dict.items()
            ])
            duration = entry['duration']
            title = entry['title']
            if self.args.resolution_on_title:
                height = self.data[url].get('height')
                title += f' [{utils.height_to_resolution(height)}]'

            res.append(
                f'{tags.INF}:{duration} {params},{title}\n{url}')

        open(self.args.output, 'w', encoding='utf-8').write('\n'.join(res))

    def parse(self):
        """Parse contents."""
        self._parse(self.args.inputs)
        logging.debug(self.data)
        self._parse(self.args.templates, is_template=True)
        logging.debug(self.data)

    def _parse(self, sources, is_template=False):
        """Parse playlist sources."""
        for source in sources:
            lines = parsers.parse_content_to_lines(source)
            source_name = os.path.splitext(os.path.basename(source))[0]

            if lines[0].startswith(tags.M3U):
                res = parsers.parse_tag_m3u(lines[0])
                if res.get('tvg-url'):
                    self.tvg_url = res.get('tvg-url')
                lines = lines[1:]

            current_item = {}
            skip = False
            for line in lines:
                line = line.strip()
                if not line:
                    continue
                if skip:
                    skip = False
                    continue
                if line.startswith(tags.INF):
                    current_item = parsers.parse_tag_inf(line)
                    current_item = utils.unify_title_and_id(current_item)
                    current_id = current_item['id']

                    params = current_item.get('params', {})
                    group = params.get('group-title', '')
                    if not skip and self.args.group_include:
                        if re.search(self.args.group_include, group):
                            logging.debug(f'Group to include: `{group}`.')
                        else:
                            skip = True
                    if not skip and self.args.group_exclude and \
                            re.search(self.args.group_exclude, group):
                        skip = True
                        logging.debug(f'Group to exclude: `{group}`.')

                    title = current_item.get('title', '')
                    if not skip and self.args.channel_include:
                        if re.search(self.args.channel_include, title):
                            logging.debug(f'Channel to include: `{title}`.')
                        else:
                            skip = True
                    if not skip and self.args.channel_exclude and \
                            re.search(self.args.channel_exclude, title):
                        skip = True
                        logging.debug(f'Channel to exclude: `{title}`.')

                else:
                    if is_template:
                        for url in self.id_url.get(current_id, []):
                            current_params = current_item['params']
                            self.data[url]['params'].update(current_params)
                            self.data[url]['title'] = current_item['title']
                    else:
                        if self.args.udpxy:
                            line = utils.convert_url_with_udpxy(
                                line, self.args.udpxy)
                        current_item['source'] = source_name
                        self.data[line] = current_item

                        if current_id not in self.id_url:
                            self.id_url[current_id] = []
                        self.id_url[current_id].append(line)

    def filter(self):
        """Filter process."""
        urls = list(self.data.keys())
        random.shuffle(urls)
        pbar = tqdm(urls, ascii=True)
        for url in pbar:
            time.sleep(self.args.interval)
            status = 'OK'
            if self.args.min_height or self.args.resolution_on_title:
                height = utils.check_stream(url, self.args.timeout)
                if height == 0:
                    self.inaccessible_urls.add(url)
                    status = 'Inaccessible'
                elif height < self.args.min_height:
                    self.poor_urls.add(url)
                    status = 'Poor Resolution'
                self.data[url]['height'] = height
            elif not utils.check_connectivity(url, self.args.timeout):
                self.inaccessible_urls.add(url)
                status = 'Inaccessible'
            pbar.write(f'{url}, {status}!')

    def __custom_sort(self, url):
        """Sort by tvg-id, resolution and title."""
        res = []
        for key in self.args.sort_keys:
            entry = self.data[url]
            if key == 'height':
                res.append(-entry.get(key, 0))
            elif key == 'title':
                res.append(entry.get(key, ''))
            elif key == 'tvg-id':
                res.append(int(entry['params'].get(key) or sys.maxsize))
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Patterns for iptvtools.

File: constants.py
Author: huxuan
Email: i(at)huxuan.org
"""
import re


PARAMS = re.compile(r'(\S+)="(.*?)"')
EXTINF = re.compile(
    r'^#EXTINF:(?P<duration>-?\d+?) ?(?P<params>.*),(?P<title>.*?)$')
EXTM3U = re.compile(r'^#EXTM3U ?(?P<params>.*)$')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Init for iptvtools.

File: __init__.py
Author: huxuan
Email: i(at)huxuan.org
"""
from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution


__version__ = '0.0.0'
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Configuration for iptvtools.

File: config.py
Author: huxuan
Email: i(at)huxuan.org
"""
import json
import os
import os.path


class MetaConfig(type):
    """Configuration for iptvtools."""

    config = {}

    @classmethod
    def init(cls, config_file):
        """Initialize configuration."""
        if os.path.isfile(config_file):
            with open(config_file) as fin:
                cls.config = json.load(fin)

    def __getattr__(cls, key):
        """Get configuration with key."""
        return cls.config.get(key, {})


class Config(metaclass=MetaConfig):  # pylint: disable=R0903
    """Configuration for iptvtools."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defaults for iptvtools.

File: constants.py
Author: huxuan
Email: i(at)huxuan.org
"""
CONFIG = 'config.json'
INPUTS = ['https://iptv-org.github.io/iptv/index.m3u']
INTERVAL = 1
LOG_LEVEL = 'INFO'
MIN_HEIGHT = 0
OUTPUT = 'iptvtools.m3u'
SORT_KEYS = ['tvg-id', 'height', 'title']
TEMPLATES = []
TIMEOUT = 10
UDPXY = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Simplified parser for m3u8 file.

File: parser.py
Author: huxuan
Email: i(at)huxuan.org
"""
import logging
import os.path
from urllib.request import urlopen

from iptvtools.constants import patterns


def parse_content_to_lines(content):
    """Universal interface to split content into lines."""
    if os.path.isfile(content):
        return _parse_from_file(content)
    return _parse_from_url(content)


def parse_tag_inf(line):
    """Parse INF content."""
    match = patterns.EXTINF.fullmatch(line)
    res = match.groupdict()
    if 'params' in res:
        res['params'] = dict(patterns.PARAMS.findall(res['params']))
    return res


def parse_tag_m3u(line):
    """Parse M3U content."""
    match = patterns.EXTM3U.fullmatch(line)
    return match.groupdict()


def _parse_from_file(filename):
    """Parse content from file."""
    logging.info(f'Retrieving playlists from file: {filename}')
    with open(filename, encoding='utf-8') as fin:
        return fin.read().splitlines()


def _parse_from_url(url):
    """Parse content from url."""
    logging.info(f'Retrieving playlists from url: {url}')
    with urlopen(url) as response:  # noqa: S310
        return response.read().decode('utf-8').splitlines()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Python packaging for iptvtools.

File: setup.py
Author: huxuan
Email: i(at)huxuan.org
"""
from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution
from setuptools import find_packages
from setuptools import setup


NAME = 'iptvtools'
AUTHOR = 'huxuan'
EMAIL = f'i+{NAME}@huxuan.org'

DESCRIPTION = (
    'A set of scripts that help to better IPTV experience.'
)

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3 :: Only',
    'Topic :: Utilities',
]

INSTALL_REQUIRES = [
    'requests',
    'tqdm',
]

KEYWORDS = [
    'iptv',
    'm3u',
    'playlist',
    'tools',
    'filter',
]

try:
    VERSION = f'v{get_distribution(NAME).version}'
except DistributionNotFound:
    VERSION = 'master'

PROJECT_URL = f'https://github.com/{AUTHOR}/{NAME}'
BASE_URL = f'{PROJECT_URL}/blob/{VERSION}'


def readme():
    """Parse README for long_description."""
    content = open('README.md').read()
    content = content.replace('README.md', f'{BASE_URL}/README.md', 1)
    content = content.replace('README-en.md', f'{BASE_URL}/README-en.md', 1)
    content = content.replace('README-zh.md', f'{BASE_URL}/README-zh.md', 1)
    return content


setup(name=NAME,
      description=DESCRIPTION,
      long_description=readme(),
      long_description_content_type='text/markdown',
      classifiers=CLASSIFIERS,
      keywords=' '.join(KEYWORDS),
      url=PROJECT_URL,
      author=AUTHOR,
      author_email=EMAIL,
      license='MIT',
      packages=find_packages(exclude=['tests']),
      use_scm_version=True,
      setup_requires=['setuptools_scm'],
      install_requires=INSTALL_REQUIRES,
      python_requires='>=3',
      entry_points={
          'console_scripts': [
              'iptv-filter=iptvtools.iptv_filter:main',
          ],
      },
      include_package_data=True)
